"""Originally Adapted from sphinxcontrib.details.directive
"""
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.nodes import NodeMatcher

from .icons import get_opticon


def setup_dropdown(app):
    app.add_node(dropdown_main, html=(visit_dropdown_main, depart_dropdown_main))
    app.add_node(dropdown_title, html=(visit_dropdown_title, depart_dropdown_title))
    app.add_directive("dropdown", DropdownDirective)
    app.add_post_transform(DropdownHtmlTransform)


class dropdown_main(nodes.Element, nodes.General):
    pass


class dropdown_title(nodes.TextElement, nodes.General):
    pass


def visit_dropdown_main(self, node):
    if node.get("opened"):
        self.body.append(self.starttag(node, "details", open="open"))
    else:
        self.body.append(self.starttag(node, "details"))


def depart_dropdown_main(self, node):
    self.body.append("</details>")


def visit_dropdown_title(self, node):
    self.body.append(self.starttag(node, "summary"))


def depart_dropdown_title(self, node):
    self.body.append("</summary>")


class DropdownDirective(SphinxDirective):
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        "container": directives.unchanged,
        "title": directives.unchanged,
        "body": directives.unchanged,
        "open": directives.flag,
        "name": directives.unchanged,
        "animate": lambda a: directives.choice(a, ("fade-in", "fade-in-slide-down")),
    }

    def run(self):

        # default classes
        classes = {
            "container_classes": ["mb-3"],
            "title_classes": [],
            "body_classes": [],
        }

        # add classes from options
        for element in ["container", "title", "body"]:
            if element not in self.options:
                continue
            value = self.options.get(element).strip()
            if value.startswith("+"):
                classes.setdefault(element + "_classes", []).extend(value[1:].split())
            else:
                classes[element + "_classes"] = value.split()

        # add animation classes
        if (
            "animate" in self.options
            and self.options["animate"] not in classes["container_classes"]
        ):
            classes["container_classes"].append(self.options["animate"])

        container = nodes.container(
            "",
            opened="open" in self.options,
            type="dropdown",
            has_title=len(self.arguments) > 0,
            **classes
        )
        if self.arguments:
            textnodes, messages = self.state.inline_text(self.arguments[0], self.lineno)
            container += nodes.paragraph(self.arguments[0], "", *textnodes)
            container += messages
        self.state.nested_parse(self.content, self.content_offset, container)
        self.add_name(container)
        return [container]


KEBAB = """\
<svg viewBox="0 0 36 24" width="36" height="16" xmlns="http://www.w3.org/2000/svg"
    class="octicon no-title" aria-hidden="true">
  <g xmlns="http://www.w3.org/2000/svg" class="jp-icon3">
    <circle cx="0" cy="12" r="6"></circle>
    <circle cx="18" cy="12" r="6"></circle>
    <circle cx="36" cy="12" r="6"></circle>
  </g>
</svg>"""


class DropdownHtmlTransform(SphinxPostTransform):
    default_priority = 200
    builders = ("html", "dirhtml", "singlehtml", "readthedocs")

    def run(self):
        matcher = NodeMatcher(nodes.container, type="dropdown")
        for node in self.document.traverse(matcher):

            open_marker = nodes.container(
                "",
                nodes.raw(
                    "", nodes.Text(get_opticon("chevron-up", size=24)), format="html"
                ),
                is_div=True,
                classes=["summary-up"],
            )
            closed_marker = nodes.container(
                "",
                nodes.raw(
                    "", nodes.Text(get_opticon("chevron-down", size=24)), format="html"
                ),
                is_div=True,
                classes=["summary-down"],
            )

            newnode = dropdown_main(
                opened=node["opened"],
                classes=["sphinx-bs", "dropdown", "card"] + node["container_classes"],
            )

            if node["has_title"]:
                title_children = node[0]
                body_children = node[1:]
            else:
                title_children = [
                    nodes.raw(
                        "...",
                        nodes.Text(
                            KEBAB
                            # Note the custom opticon here has thicker dots
                            # get_opticon("kebab-horizontal", classes="no-title",
                            # size=24)
                        ),
                        format="html",
                    )
                ]
                body_children = node

            newnode += dropdown_title(
                "",
                "",
                *title_children,
                closed_marker,
                open_marker,
                classes=["summary-title", "card-header"] + node["title_classes"]
            )
            body_node = nodes.container(
                "",
                *body_children,
                is_div=True,
                classes=["summary-content", "card-body"] + node["body_classes"]
            )
            for para in body_node.traverse(nodes.paragraph):
                para["classes"] = ([] if "classes" in para else para["classes"]) + [
                    "card-text"
                ]
            newnode += body_node
            # newnode += open_marker
            node.replace_self(newnode)



""""A sphinx extension to add a ``panels`` directive."""
import hashlib
from pathlib import Path

try:
    import importlib.resources as resources
except ImportError:
    # python < 3.7
    import importlib_resources as resources

from docutils import nodes
from docutils.parsers.rst import directives, Directive
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.util.logging import getLogger

from .button import setup_link_button
from .dropdown import setup_dropdown
from .panels import setup_panels
from .tabs import setup_tabs
from .icons import setup_icons

from . import _css as css_module

__version__ = "0.5.1"

LOGGER = getLogger(__name__)


def get_default_css_variables():
    return {
        "tabs-color-label-active": "hsla(231, 99%, 66%, 1)",
        "tabs-color-label-inactive": "rgba(178, 206, 245, 0.62)",
        "tabs-color-overline": "rgb(207, 236, 238)",
        "tabs-color-underline": "rgb(207, 236, 238)",
        "tabs-size-label": "1rem",
    }


def update_css(app: Sphinx):
    """Compile SCSS to the build directory."""
    # merge user CSS variables with defaults
    css_variables = get_default_css_variables()
    for key, value in app.config.panels_css_variables.items():
        if key not in css_variables:
            LOGGER.warning(f"key in 'panels_css_variables' is not recognised: {key}")
        else:
            css_variables[key] = value

    # reset css changed attribute
    app.env.panels_css_changed = False

    # setup up new static path in output dir
    static_path = (Path(app.outdir) / "_panels_static").absolute()
    static_path.mkdir(exist_ok=True)
    app.config.html_static_path.append(str(static_path))

    # record current resources
    old_resources = {path.name for path in static_path.glob("*") if path.is_file()}

    # Add core CSS
    css_files = [r for r in resources.contents(css_module) if r.endswith(".css")]
    if not app.config.panels_add_boostrap_css:
        css_files = [name for name in css_files if "bootstrap" not in name]
    for filename in css_files:
        app.add_css_file(filename)
        if not (static_path / filename).exists():
            content = resources.read_text(css_module, filename)
            (static_path / filename).write_text(content)
            app.env.panels_css_changed = True
        if filename in old_resources:
            old_resources.remove(filename)

    # add variables CSS file
    css_lines = [":root {"]
    for name, value in css_variables.items():
        css_lines.append(f"--{name}: {value};")
    css_lines.append("}")
    css_str = "\n".join(css_lines)
    css_variables_name = (
        f"panels-variables.{hashlib.md5(css_str.encode('utf8')).hexdigest()}.css"
    )
    app.add_css_file(css_variables_name)
    if not (static_path / css_variables_name).exists():
        (static_path / css_variables_name).write_text(css_str)
        app.env.panels_css_changed = True
    if css_variables_name in old_resources:
        old_resources.remove(css_variables_name)

    # remove old resources
    for name in old_resources:
        for path in Path(app.outdir).glob(f"**/{name}"):
            path.unlink()


def update_css_links(app: Sphinx, env: BuildEnvironment):
    """If CSS has changed, all files must be re-written,
    to include the correct stylesheets.
    """
    if env.panels_css_changed and app.config.panels_dev_mode:
        LOGGER.debug("sphinx-panels CSS changed; re-writing all files")
        return list(env.all_docs.keys())


class Div(Directive):
    """Same as the ``container`` directive,
    but does not add the ``container`` class in HTML outputs,
    which can interfere with Bootstrap CSS.
    """

    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {"name": directives.unchanged}
    has_content = True

    def run(self):
        self.assert_has_content()
        text = "\n".join(self.content)
        try:
            if self.arguments:
                classes = directives.class_option(self.arguments[0])
            else:
                classes = []
        except ValueError:
            raise self.error(
                'Invalid class attribute value for "%s" directive: "%s".'
                % (self.name, self.arguments[0])
            )
        node = nodes.container(text, is_div=True)
        node["classes"].extend(classes)
        self.add_name(node)
        self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


def visit_container(self, node: nodes.Node):
    classes = "docutils container"
    if node.get("is_div", False):
        # we don't want the CSS for container for these nodes
        classes = "docutils"
    self.body.append(self.starttag(node, "div", CLASS=classes))


def depart_container(self, node: nodes.Node):
    self.body.append("</div>\n")


def setup(app: Sphinx):
    app.add_directive("div", Div)
    app.add_config_value("panels_add_boostrap_css", True, "env")
    app.add_config_value("panels_css_variables", {}, "env")
    app.add_config_value("panels_dev_mode", False, "env")
    app.connect("builder-inited", update_css)
    app.connect("env-updated", update_css_links)
    # we override container html visitors, to stop the default behaviour
    # of adding the `container` class to all nodes.container
    app.add_node(
        nodes.container, override=True, html=(visit_container, depart_container)
    )

    setup_panels(app)
    setup_link_button(app)
    setup_dropdown(app)
    setup_tabs(app)
    setup_icons(app)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from urllib.parse import unquote

from docutils import nodes
from docutils.utils import unescape
from docutils.parsers.rst import directives
from sphinx import addnodes
from sphinx.util.docutils import SphinxDirective

from .utils import string_to_func_inputs


def setup_link_button(app):
    app.add_directive("link-button", LinkButton)
    # TODO hide badges in non-HTML?
    app.add_role("badge", badge_role)
    app.add_role("link-badge", link_badge_role)


def create_ref_node(link_type, uri, text, tooltip):
    innernode = nodes.inline(text, text)
    if link_type == "ref":
        ref_node = addnodes.pending_xref(
            reftarget=unquote(uri),
            reftype="any",
            # refdoc=self.env.docname,
            refdomain="",
            refexplicit=True,
            refwarn=True,
        )
        innernode["classes"] = ["xref", "any"]
        # if tooltip:
        #     ref_node["reftitle"] = tooltip
        #     ref_node["title"] = tooltip
        # TODO this doesn't work
    else:
        ref_node = nodes.reference()
        ref_node["refuri"] = uri
        if tooltip:
            ref_node["reftitle"] = tooltip
    ref_node += innernode
    return ref_node


class LinkButton(SphinxDirective):
    """A directive to turn a link into a button."""

    has_content = False
    required_arguments = 1
    final_argument_whitespace = True
    option_spec = {
        "type": lambda arg: directives.choice(arg, ("url", "ref")),
        "text": directives.unchanged,
        "tooltip": directives.unchanged,
        "classes": directives.unchanged,
    }

    def run(self):

        uri = self.arguments[0]
        link_type = self.options.get("type", "url")

        text = self.options.get("text", uri)

        ref_node = create_ref_node(
            link_type, uri, text, self.options.get("tooltip", None)
        )
        self.set_source_info(ref_node)
        ref_node["classes"] = ["sphinx-bs", "btn", "text-wrap"] + self.options.get(
            "classes", ""
        ).split()

        # sphinx requires that a reference be inside a block element
        container = nodes.paragraph()
        container += ref_node

        return [container]


def get_badge_inputs(text, cls: str = ""):
    return text, cls.split()


def badge_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    try:
        args, kwargs = string_to_func_inputs(text)
        text, classes = get_badge_inputs(*args, **kwargs)
    except Exception as err:
        msg = inliner.reporter.error(f"badge input is invalid: {err}", line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    node = nodes.inline(
        rawtext, unescape(text), classes=["sphinx-bs", "badge"] + classes
    )
    # textnodes, messages = inliner.parse(text, lineno, node, memo)
    # TODO this would require the memo with reporter, document and language
    return [node], []


def get_link_badge_inputs(link, text=None, type="link", cls: str = "", tooltip=None):
    return link, text or link, type, cls.split(), tooltip


def link_badge_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    try:
        args, kwargs = string_to_func_inputs(text)
        uri, text, link_type, classes, tooltip = get_link_badge_inputs(*args, **kwargs)
    except Exception as err:
        msg = inliner.reporter.error(f"badge input is invalid: {err}", line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    ref_node = create_ref_node(link_type, uri, text, tooltip)
    if lineno is not None:
        ref_node.source, ref_node.line = inliner.reporter.get_source_and_line(lineno)
    ref_node["classes"] = ["sphinx-bs", "badge"] + classes
    return [ref_node], []



from functools import lru_cache
import json
from pathlib import Path

from docutils import nodes

from .utils import string_to_func_inputs


OPTICON_VERSION = "0.0.0-dd899ea"

OPTICON_CSS = """\
.octicon {
  display: inline-block;
  vertical-align: text-top;
  fill: currentColor;
}"""


@lru_cache(1)
def get_opticon_data():
    path = Path(__file__).parent.joinpath("data", "opticons.json")
    return json.loads(path.read_text())


def list_opticons():
    return list(get_opticon_data().keys())


def get_opticon(
    name: str,
    classes: str = None,
    width: int = None,
    height: int = None,
    aria_label: str = None,
    size: int = 16,
):
    assert size in [16, 24], "size must be 16 or 24"
    try:
        data = get_opticon_data()[name]
    except KeyError:
        raise KeyError(f"Unrecognised opticon: {name}")

    content = data["heights"][str(size)]["path"]
    options = {
        "version": "1.1",
        "width": data["heights"][str(size)]["width"],
        "height": int(size),
        "class": f"octicon octicon-{name}",
    }

    if width is not None or height is not None:
        if width is None:
            width = round((int(height) * options["width"]) / options["height"], 2)
        if height is None:
            height = round((int(width) * options["height"]) / options["width"], 2)
        options["width"] = width
        options["height"] = height

    options["viewBox"] = f'0 0 {options["width"]} {options["height"]}'

    if classes is not None:
        options["class"] += " " + classes.strip()

    if aria_label is not None:
        options["aria-label"] = aria_label
        options["role"] = "img"
    else:
        options["aria-hidden"] = "true"

    opt_string = " ".join(f'{k}="{v}"' for k, v in options.items())
    return f"<svg {opt_string}>{content}</svg>"


def opticon_role(
    role, rawtext: str, text: str, lineno, inliner, options={}, content=[]
):
    try:
        args, kwargs = string_to_func_inputs(text)
        svg = get_opticon(*args, **kwargs)
    except Exception as err:
        msg = inliner.reporter.error(f"Opticon input is invalid: {err}", line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    node = nodes.raw("", nodes.Text(svg), format="html")
    return [node], []


class fontawesome(nodes.Element, nodes.General):
    pass


def create_fa_node(name, classes: str = None, style="fa"):
    assert style.startswith("fa"), "style must be a valid prefix, e.g. fa, fas, etc"
    return fontawesome(
        icon_name=name,
        classes=[style, f"fa-{name}"] + (classes.split() if classes else []),
    )


def fontawesome_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    try:
        args, kwargs = string_to_func_inputs(text)
        node = create_fa_node(*args, **kwargs)
    except Exception as err:
        msg = inliner.reporter.error(
            f"FontAwesome input is invalid: {err}", line=lineno
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    return [node], []


def visit_fontawesome_html(self, node):
    self.body.append(self.starttag(node, "span", ""))


def depart_fontawesome_html(self, node):
    self.body.append("</span>")


def visit_fontawesome_latex(self, node):
    if self.config.panels_add_fontawesome_latex:
        self.body.append(f"\\faicon{{{node['icon_name']}}}")
    raise nodes.SkipNode


def add_fontawesome_pkg(app, config):
    if app.config.panels_add_fontawesome_latex:
        app.add_latex_package("fontawesome")


def setup_icons(app):
    app.add_role("opticon", opticon_role)
    app.add_role("fa", fontawesome_role)

    app.add_config_value("panels_add_fontawesome_latex", False, "env")
    app.connect("config-inited", add_fontawesome_pkg)
    app.add_node(
        fontawesome,
        html=(visit_fontawesome_html, depart_fontawesome_html),
        latex=(visit_fontawesome_latex, None),
        text=(None, None),
        man=(None, None),
        texinfo=(None, None),
    )



from uuid import uuid4
from typing import Optional

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.docutils import SphinxDirective
from sphinx.util.logging import getLogger
from sphinx.util.nodes import NodeMatcher

LOGGER = getLogger(__name__)


def setup_tabs(app):
    app.add_directive("tabbed", TabbedDirective)
    app.add_post_transform(TabbedHtmlTransform)
    app.add_node(tabbed_input, html=(visit_tabbed_input, depart_tabbed_input))
    app.add_node(tabbed_label, html=(visit_tabbed_label, depart_tabbed_label))


class tabbed_input(nodes.Element, nodes.General):
    pass


class tabbed_label(nodes.TextElement, nodes.General):
    pass


def visit_tabbed_input(self, node):
    attributes = {"ids": [node["id"]], "type": node["type"], "name": node["set_id"]}
    if node["checked"]:
        attributes["checked"] = "checked"
    self.body.append(self.starttag(node, "input", **attributes))


def depart_tabbed_input(self, node):
    self.body.append("</input>")


def visit_tabbed_label(self, node):
    attributes = {"for": node["input_id"]}
    self.body.append(self.starttag(node, "label", **attributes))


def depart_tabbed_label(self, node):
    self.body.append("</label>")


class TabbedDirective(SphinxDirective):
    """CSS-based tabs."""

    required_arguments = 1
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        "new-group": directives.flag,
        "selected": directives.flag,
        "name": directives.unchanged,
        "class-label": directives.class_option,
        "class-content": directives.class_option,
    }

    def run(self):
        self.assert_has_content()

        container = nodes.container(
            "",
            type="tabbed",
            new_group="new-group" in self.options,
            selected="selected" in self.options,
            classes=["tabbed-container"],
        )
        self.set_source_info(container)

        # add label as a rubric (to degrade nicely for non-html outputs)
        textnodes, messages = self.state.inline_text(self.arguments[0], self.lineno)
        label = nodes.rubric(self.arguments[0], *textnodes, classes=["tabbed-label"])
        label["classes"] += self.options.get("class-label", [])
        self.add_name(label)
        container += label

        # add content
        content = nodes.container("", is_div=True, classes=["tabbed-content"])
        content["classes"] += self.options.get("class-content", [])
        self.state.nested_parse(self.content, self.content_offset, content)

        container += content

        return [container]


class TabSet:
    def __init__(self, node):
        self._nodes = [node]

    def is_next(self, node):
        if self.parent != node.parent:
            return False
        if node.parent.index(node) != (self.indices[-1] + 1):
            return False
        return True

    def append(self, node):
        assert self.is_next(node)
        self._nodes.append(node)

    @property
    def parent(self) -> int:
        return self._nodes[0].parent

    @property
    def nodes(self) -> int:
        return self._nodes[:]

    @property
    def indices(self) -> int:
        return [n.parent.index(n) for n in self._nodes]


class TabbedHtmlTransform(SphinxPostTransform):
    default_priority = 200
    builders = ("html", "dirhtml", "singlehtml", "readthedocs")

    def get_unique_key(self):
        return str(uuid4())

    def run(self):
        matcher = NodeMatcher(nodes.container, type="tabbed")
        tab_set = None
        for node in self.document.traverse(matcher):  # type: nodes.container
            if tab_set is None:
                tab_set = TabSet(node)
            elif node["new_group"]:
                self.render_tab_set(tab_set)
                tab_set = TabSet(node)
            elif tab_set.is_next(node):
                tab_set.append(node)
            else:
                self.render_tab_set(tab_set)
                tab_set = TabSet(node)
        self.render_tab_set(tab_set)

    def render_tab_set(self, tab_set: Optional[TabSet]):

        if tab_set is None:
            return

        container = nodes.container("", is_div=True, classes=["tabbed-set"])
        container.parent = tab_set.parent
        set_identity = self.get_unique_key()

        # get the first selected node
        selected_idx = None
        for idx, tab in enumerate(tab_set.nodes):
            if tab["selected"]:
                if selected_idx is None:
                    selected_idx = idx
                else:
                    LOGGER.warning("multiple selected tabbed directives", location=tab)
        selected_idx = 0 if selected_idx is None else selected_idx

        for idx, tab in enumerate(tab_set.nodes):
            # TODO warn and continue if incorrect children
            title, content = tab.children
            # input <input checked="checked" id="id" type="radio">
            identity = self.get_unique_key()
            input_node = tabbed_input(
                "",
                id=identity,
                set_id=set_identity,
                type="radio",
                checked=(idx == selected_idx),
            )
            input_node.source, input_node.line = tab.source, tab.line
            container += input_node
            # label <label for="id">Title</label>
            # TODO this actually has to be text only
            label = tabbed_label("", *title.children, input_id=identity)
            label["classes"] = title["classes"]
            container += label
            input_node.source, input_node.line = tab.source, tab.line
            # content
            container += content

        # replace all nodes
        tab_set.parent.children = (
            tab_set.parent.children[: tab_set.indices[0]]
            + [container]
            + tab_set.parent.children[tab_set.indices[-1] + 1 :]
        )



from ast import literal_eval
import re

REGEX = re.compile(
    r'\s*(?:(?P<key>[a-zA-Z0-9_]+)\s*\=)?\s*(?P<value>".*"|[^,]+)\s*(?:,|$)'
)


def eval_literal(string):
    try:
        value = literal_eval(string)
    except Exception:
        value = string
    return value


def string_to_func_inputs(text):
    args = []
    kwargs = {}
    for key, value in REGEX.findall(text):
        if key:
            kwargs[key.strip()] = eval_literal(value.strip())
        else:
            args.append(eval_literal(value.strip()))
    return args, kwargs



import re

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective

DEFAULT_CONTAINER = "container pb-4"
DEFAULT_COLUMN = "col-lg-6 col-md-6 col-sm-6 col-xs-12 p-2"
DEFAULT_CARD = "shadow"

RE_OPTIONS = re.compile(
    r"\:(column|card|body|header|footer|"
    r"img-top|img-bottom|img-top-cls|img-bottom-cls)\:\s*(\+?)\s*(.*)"
)


def setup_panels(app):
    app.add_directive("panels", Panels)
    app.add_config_value(
        "panels_delimiters", (r"^\-{3,}$", r"^\^{3,}$", r"^\+{3,}$"), "env"
    )
    app.connect("config-inited", validate_config)


def parse_panels(
    content,
    content_offset,
    default_classes,
    panel_regex=None,
    head_regex=None,
    foot_regex=None,
):
    """split a block of content into panels.

    example::

        ---
        header
        ===
        body
        ...
        footer
        ---
        next panel

    """
    panel_regex = panel_regex or re.compile(r"^\-{3,}$")
    head_regex = head_regex or re.compile(r"^\^{3,}$")
    foot_regex = foot_regex or re.compile(r"^\+{3,}$")

    if isinstance(content, str):
        content = content.splitlines()

    panel_blocks = []
    start_line = 0
    header_split = footer_split = None
    for i, line in enumerate(content):
        if panel_regex.match(line.strip()):
            if i != 0:
                panel_blocks.append(
                    parse_single_panel(
                        content[start_line:i],
                        start_line,
                        header_split,
                        footer_split,
                        content_offset,
                        default_classes,
                    )
                )
            start_line = i + 1
            header_split = footer_split = None
        if head_regex.match(line.strip()) and footer_split is None:
            header_split = i - start_line
        if foot_regex.match(line.strip()):
            footer_split = i - start_line
        # TODO warn if multiple header_split or footer_split
        # TODO assert header_split is before footer_split
    try:
        panel_blocks.append(
            parse_single_panel(
                content[start_line:],
                start_line,
                header_split,
                footer_split,
                content_offset,
                default_classes,
            )
        )
    except IndexError:
        pass
    return panel_blocks


def parse_single_panel(
    content, offset, header_split, footer_split, content_offset, default_classes
):
    """parse each panel data to dict."""
    output = {}
    body_start = 0
    body_end = len(content)

    # parse the classes required for this panel, and top/bottom images
    classes = default_classes.copy()
    for opt_offset, line in enumerate(content):
        opt_match = RE_OPTIONS.match(line)
        if not opt_match:
            break
        body_start += 1
        if opt_match.group(1) in ["img-top", "img-bottom"]:
            output[opt_match.group(1)] = opt_match.group(3)
            continue
        if opt_match.group(2) == "+":
            classes[opt_match.group(1)] = (
                classes.get(opt_match.group(1), []) + opt_match.group(3).split()
            )
        else:
            classes[opt_match.group(1)] = opt_match.group(3).split()

    if classes:
        output["classes"] = classes

    if header_split is not None:
        header_content = content[opt_offset:header_split]
        header_offset = content_offset + offset + opt_offset
        body_start = header_split + 1
        output["header"] = (header_content, header_offset)

    if footer_split is not None:
        footer_content = content[footer_split + 1 :]
        footer_offset = content_offset + offset + footer_split
        body_end = footer_split
        output["footer"] = (footer_content, footer_offset)

    body_content = content[body_start:body_end]
    body_offset = content_offset + offset + body_start
    output["body"] = (body_content, body_offset)
    return output


def add_child_classes(node):
    """Add classes to specific child nodes."""
    for para in node.traverse(nodes.paragraph):
        para["classes"] = ([] if "classes" in para else para["classes"]) + ["card-text"]
    for title in node.traverse(nodes.title):
        title["classes"] = ([] if "classes" in title else title["classes"]) + [
            "card-title"
        ]


class Panels(SphinxDirective):
    """Two Column Panels."""

    has_content = True
    option_spec = {
        "container": directives.unchanged,
        "column": directives.unchanged,
        "card": directives.unchanged,
        "body": directives.unchanged,
        "header": directives.unchanged,
        "footer": directives.unchanged,
        "img-top-cls": directives.unchanged,
        "img-bottom-cls": directives.unchanged,
    }

    def run(self):
        default_classes = {
            "container": DEFAULT_CONTAINER.split(),
            "column": DEFAULT_COLUMN.split(),
            "card": DEFAULT_CARD.split(),
            "body": [],
            "header": [],
            "footer": [],
            "img-top-cls": [],
            "img-bottom-cls": [],
        }

        # set classes from the directive options
        for key, value in default_classes.items():
            if key not in self.options:
                continue
            option_value = self.options[key].strip()
            if option_value.startswith("+"):
                default_classes[key] += option_value[1:].split()
            else:
                default_classes[key] = option_value.split()

        # split the block into panels
        panel_blocks = parse_panels(
            self.content,
            self.content_offset,
            default_classes,
            panel_regex=self.env.app.config.panels_delimiters[0],
            head_regex=self.env.app.config.panels_delimiters[1],
            foot_regex=self.env.app.config.panels_delimiters[2],
        )

        # set the top-level containers
        parent = nodes.container(
            is_div=True, classes=["sphinx-bs"] + default_classes["container"]
        )
        rows = nodes.container(is_div=True, classes=["row"])
        parent += rows

        for data in panel_blocks:

            classes = data["classes"]

            column = nodes.container(
                is_div=True, classes=["d-flex"] + classes["column"]
            )
            rows += column
            card = nodes.container(
                is_div=True, classes=["card", "w-100"] + classes["card"]
            )
            column += card

            if "img-top" in data:
                image_top = nodes.image(
                    "",
                    uri=directives.uri(data["img-top"]),
                    alt="img-top",
                    classes=["card-img-top"] + classes["img-top-cls"],
                )
                self.add_name(image_top)
                card += image_top

            if "header" in data:
                header = nodes.container(
                    is_div=True, classes=["card-header"] + classes["header"]
                )
                card += header

                header_content, header_offset = data["header"]
                self.state.nested_parse(header_content, header_offset, header)
                add_child_classes(header)

            body = nodes.container(is_div=True, classes=["card-body"] + classes["body"])
            card += body

            body_content, body_offset = data["body"]
            self.state.nested_parse(body_content, body_offset, body)
            add_child_classes(body)

            if "footer" in data:
                footer = nodes.container(
                    is_div=True, classes=["card-footer"] + classes["footer"]
                )
                card += footer

                footer_content, footer_offset = data["footer"]
                self.state.nested_parse(footer_content, footer_offset, footer)
                add_child_classes(footer)

            if "img-bottom" in data:
                image_top = nodes.image(
                    "",
                    uri=directives.uri(data["img-bottom"]),
                    alt="img-bottom",
                    classes=["card-img-bottom"] + classes["img-bottom-cls"],
                )
                self.add_name(image_top)
                card += image_top

        return [parent]


def validate_config(app, config):
    if len(app.config.panels_delimiters) != 3:
        raise AssertionError(
            "panels_delimiters config must be of form: (header, body, footer)"
        )
    if len(set(app.config.panels_delimiters)) != 3:
        raise AssertionError("panels_delimiters config must contain unique values")
    try:
        app.config.panels_delimiters = tuple(
            [re.compile(s) for s in app.config.panels_delimiters]
        )
    except Exception as err:
        raise AssertionError(
            "panels_delimiters config must contain only compilable regexes: {}".format(
                err
            )
        )



from pathlib import Path
from setuptools import setup, find_packages

version = [
    line
    for line in Path("sphinx_panels/__init__.py").read_text().split("\n")
    if "__version__" in line
]
version = version[0].split(" = ")[-1].strip('"')

with open("./README.md", "r") as ff:
    readme_text = ff.read()

setup(
    name="sphinx-panels",
    version=version,
    description="A sphinx extension for creating panels in a grid layout.",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    author="Chris Sewell",
    author_email="chrisj_sewell@hotmail.com",
    url="https://github.com/executablebooks/sphinx-panels",
    project_urls={"Documentation": "https://sphinx-panels.readthedocs.io"},
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "docutils",
        "sphinx>=2,<4",
        'importlib-resources~=3.0.0; python_version < "3.7"',
    ],
    extras_require={
        "themes": [
            "sphinx-rtd-theme",
            "pydata-sphinx-theme~=0.4.0",
            "sphinx-book-theme~=0.0.36",
            "myst-parser~=0.12.9",
        ],
        "code_style": ["pre-commit~=2.7.0"],
        "testing": ["pytest~=6.0.1", "pytest-regressions~=2.0.1"],
        "live-dev": ["sphinx-autobuild", "web-compile~=0.2.0"],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Sphinx :: Extension",
    ],
    keywords="sphinx html bootstrap grid card dropdown button badge",
)
